@@ -309,7 +309,7 @@ channel_ready (TpChannel *channel_proxy,
         TP_IFACE_QUARK_CHANNEL_INTERFACE_MEDIA_SIGNALLING))
     {
       GError e = { TP_ERRORS, TP_ERROR_NOT_IMPLEMENTED,
-        "Stream Engine was passed a channel that does not implement "
+        "Channel does not implement "
         TP_IFACE_CHANNEL_INTERFACE_MEDIA_SIGNALLING };
 
       g_message ("%s", e.message);
@@ -408,9 +408,21 @@ tf_channel_dispose (GObject *object)
       for (i = 0; i < self->priv->sessions->len; i++)
         {
           GObject *obj = g_ptr_array_index (self->priv->sessions, i);
+          FsConference *conf = NULL;
+          FsParticipant *part = NULL;
 
           g_signal_handlers_disconnect_by_func (obj, new_stream_cb, self);
 
+          g_object_get (obj,
+              "farsight-conference", &conf,
+              "farsight-participant", &part,
+              NULL);
+
+          g_signal_emit (self, signals[SESSION_INVALIDATED], 0, conf, part);
+
+          g_object_unref (conf);
+          g_object_unref (part);
+
           g_object_unref (g_ptr_array_index (self->priv->sessions, i));
         }
 
@@ -636,6 +648,16 @@ stream_closed_cb (TfStream *stream,
   g_ptr_array_index (self->priv->streams, stream_id) = NULL;
 }
 
+static void
+stream_created_cb (TfStream *stream, gpointer user_data)
+{
+  TfChannel *self = user_data;
+
+  g_signal_emit (self, signals[STREAM_CREATED], 0, stream);
+
+  _tf_stream_try_sending_codecs (stream);
+}
+
 static void
 new_stream_cb (TfSession *session,
     gchar *object_path,
@@ -650,7 +672,6 @@ new_stream_cb (TfSession *session,
   FsParticipant *fs_participant;
   TpProxy *channel_as_proxy = (TpProxy *) self->priv->channel_proxy;
   TpMediaStreamHandler *proxy;
-  GError *error = NULL;
   GList *local_codec_config = NULL;
 
   proxy = tp_media_stream_handler_new (channel_as_proxy->dbus_daemon,
@@ -658,8 +679,11 @@ new_stream_cb (TfSession *session,
 
   if (proxy == NULL)
     {
-      g_warning ("failed to construct TpMediaStreamHandler: bad object path "
-          "'%s'?", object_path);
+      gchar *str = g_strdup_printf ("failed to construct TpMediaStreamHandler:"
+          " bad object path '%s'?", object_path);
+      g_warning ("%s", str);
+      tf_channel_error (self, 0, str);
+      g_free (str);
       return;
     }
 
@@ -676,14 +700,14 @@ new_stream_cb (TfSession *session,
 
   stream = _tf_stream_new ((gpointer) self, fs_conference,
       fs_participant, proxy, stream_id, media_type, direction,
-      &self->priv->nat_props, local_codec_config, &error);
+      &self->priv->nat_props, local_codec_config,
+      stream_created_cb);
 
   fs_codec_list_destroy (local_codec_config);
 
   if (!stream)
     {
-      g_warning ("Error creating stream: %s", error->message);
-      g_clear_error (&error);
+      tf_channel_error (self, 0, "Error creating stream");
       return;
     }
 
@@ -710,10 +734,6 @@ new_stream_cb (TfSession *session,
   g_ptr_array_index (self->priv->streams, stream_id) = stream;
   g_signal_connect (stream, "closed", G_CALLBACK (stream_closed_cb),
       self);
-
-  g_signal_emit (self, signals[STREAM_CREATED], 0, stream);
-
-  _tf_stream_try_sending_codecs (stream);
 }
 
 static void
@@ -723,6 +743,8 @@ session_invalidated_cb (TfSession *session, gpointer user_data)
   FsConference *conf = NULL;
   FsParticipant *part = NULL;
 
+  g_signal_handlers_disconnect_by_func (session, new_stream_cb, self);
+
   g_object_get (session,
       "farsight-conference", &conf,
       "farsight-participant", &part,
@@ -732,6 +754,9 @@ session_invalidated_cb (TfSession *session, gpointer user_data)
 
   g_object_unref (conf);
   g_object_unref (part);
+
+  g_object_unref (session);
+  g_ptr_array_remove_fast (self->priv->sessions, session);
 }
 
 static void
@@ -755,9 +780,13 @@ add_session (TfChannel *self,
 
   if (proxy == NULL)
     {
-      g_warning ("failed to construct TpMediaSessionHandler: %s",
+      gchar *str = g_strdup_printf ("failed to construct TpMediaSessionHandler:"
+          " %s",
           error->message);
       g_error_free (error);
+      g_warning ("%s", str);
+      tf_channel_error (self, 0, str);
+      g_free (str);
       return;
     }
 
@@ -765,8 +794,12 @@ add_session (TfChannel *self,
 
   if (session == NULL)
     {
-      g_warning ("failed to create session: %s", error->message);
+      gchar *str = g_strdup_printf("failed to create session: %s",
+          error->message);
       g_error_free (error);
+      g_warning ("%s", str);
+      tf_channel_error (self, 0, str);
+      g_free (str);
       return;
     }
 
@@ -862,14 +895,14 @@ get_session_handlers_reply (TpChannel *channel_proxy G_GNUC_UNUSED,
   TfChannel *self = TF_CHANNEL (weak_object);
   guint i;
 
-  self->priv->sessions = g_ptr_array_sized_new (session_handlers->len);
-
   if (error)
     {
       g_critical ("Error calling GetSessionHandlers: %s", error->message);
       return;
     }
 
+  self->priv->sessions = g_ptr_array_sized_new (session_handlers->len);
+
   if (session_handlers->len == 0)
     {
       g_debug ("GetSessionHandlers returned 0 sessions");
