@@ -92,6 +92,8 @@ struct _TfStreamPrivate
   gboolean send_local_codecs;
   gboolean send_supported_codecs;
 
+  guint tos;
+
   NewStreamCreatedCb *new_stream_created_cb;
 };
 
@@ -119,7 +121,8 @@ enum
   PROP_DIRECTION,
   PROP_NAT_PROPERTIES,
   PROP_SINK_PAD,
-  PROP_LOCAL_PREFERENCES
+  PROP_LOCAL_PREFERENCES,
+  PROP_TOS
 };
 
 static void get_all_properties_cb (TpProxy *proxy,
@@ -239,6 +242,12 @@ tf_stream_get_property (GObject    *object,
     case PROP_LOCAL_PREFERENCES:
       g_value_set_boxed (value, self->priv->local_preferences);
       break;
+    case PROP_TOS:
+      if (self->priv->fs_session)
+        g_object_get_property (G_OBJECT (self->priv->fs_session), "tos", value);
+      else
+        g_value_set_uint (value, self->priv->tos);
+      break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
       break;
@@ -286,13 +295,18 @@ tf_stream_set_property (GObject      *object,
     case PROP_LOCAL_PREFERENCES:
       self->priv->local_preferences = g_value_dup_boxed (value);
       break;
+    case PROP_TOS:
+      self->priv->tos = g_value_get_uint (value);
+      if (self->priv->fs_session)
+        g_object_set_property (G_OBJECT (self->priv->fs_session), "tos", value);
+      break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
       break;
     }
 }
 
-#define MAX_STREAM_TRANS_PARAMS 6
+#define MAX_STREAM_TRANS_PARAMS 7
 
 static GObject *
 tf_stream_constructor (GType type,
@@ -480,6 +494,13 @@ tf_stream_class_init (TfStreamClass *klass)
           G_PARAM_READWRITE |
           G_PARAM_STATIC_STRINGS));
 
+  g_object_class_install_property (object_class, PROP_TOS,
+      g_param_spec_uint ("tos",
+          "IP Type of Service",
+          "The IP Type of Service to set on sent packets",
+          0, 255, 0,
+          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
+
   /**
    * TfStream::closed:
    * @stream: the stream that has been closed
@@ -589,11 +610,15 @@ get_all_properties_cb (TpProxy *proxy,
   GList *preferred_local_candidates = NULL;
   GParameter params[MAX_STREAM_TRANS_PARAMS];
   const gchar *nat_traversal = NULL;
-  GPtrArray *stun_servers;
+  GPtrArray *stun_servers = NULL;
   gboolean got_stun = FALSE;
-  GPtrArray *dbus_relay_info;
+  GPtrArray *dbus_relay_info = NULL;
+  gboolean created_locally = TRUE;
+  gboolean valid = FALSE;
 
-  if (dbus_error)
+  if (dbus_error &&
+      !(dbus_error->domain == DBUS_GERROR &&
+          dbus_error->code == DBUS_GERROR_UNKNOWN_METHOD))
     {
       tf_stream_error (stream, 0, dbus_error->message);
       return;
@@ -635,7 +660,8 @@ get_all_properties_cb (TpProxy *proxy,
 
   memset (params, 0, sizeof(GParameter) * MAX_STREAM_TRANS_PARAMS);
 
-  nat_traversal = tp_asv_get_string (out_Properties, "NATTraversal");
+  if (out_Properties)
+    nat_traversal = tp_asv_get_string (out_Properties, "NATTraversal");
   if (!nat_traversal && stream->priv->nat_props)
     nat_traversal = stream->priv->nat_props->nat_traversal;
 
@@ -685,8 +711,9 @@ get_all_properties_cb (TpProxy *proxy,
     }
 
   /* FIXME: use correct macro when available */
-  stun_servers = tp_asv_get_boxed (out_Properties, "STUNServers",
-      tp_type_dbus_array_su ());
+  if (out_Properties)
+    stun_servers = tp_asv_get_boxed (out_Properties, "STUNServers",
+        tp_type_dbus_array_su ());
 
   if (stun_servers && stun_servers->len)
     {
@@ -753,8 +780,9 @@ get_all_properties_cb (TpProxy *proxy,
         }
     }
 
-  dbus_relay_info = tp_asv_get_boxed (out_Properties, "RelayInfo",
-      TP_ARRAY_TYPE_STRING_VARIANT_MAP_LIST);
+  if (out_Properties)
+    dbus_relay_info = tp_asv_get_boxed (out_Properties, "RelayInfo",
+        TP_ARRAY_TYPE_STRING_VARIANT_MAP_LIST);
 
   if (dbus_relay_info && dbus_relay_info->len)
     {
@@ -820,6 +848,19 @@ get_all_properties_cb (TpProxy *proxy,
       g_value_array_free (fs_relay_info);
     }
 
+  if (out_Properties)
+    {
+      created_locally = tp_asv_get_boolean (out_Properties, "CreatedLocally",
+          &valid);
+      if (valid)
+        {
+          params[n_args].name = "controlling-mode";
+          g_value_init (&params[n_args].value, G_TYPE_BOOLEAN);
+          g_value_set_boolean (&params[n_args].value, created_locally);
+          n_args++;
+        }
+    }
+
   if (preferred_local_candidates)
     {
       params[n_args].name = "preferred-local-candidates";
@@ -842,6 +883,9 @@ get_all_properties_cb (TpProxy *proxy,
       return;
     }
 
+  if (stream->priv->tos)
+    g_object_set (stream->priv->fs_session, "tos", stream->priv->tos, NULL);
+
   stream->priv->fs_stream = fs_session_new_stream (stream->priv->fs_session,
       stream->priv->fs_participant,
       FS_DIRECTION_NONE,
@@ -863,21 +907,25 @@ get_all_properties_cb (TpProxy *proxy,
             stream->priv->local_preferences,
             &myerror))
       {
-        tf_stream_error (stream, 0, myerror->message);
-        WARNING (stream, "Error setting codec preferences: %s",
-            myerror->message);
+        if (!(myerror->domain == FS_ERROR &&
+                myerror->code == FS_ERROR_NOT_IMPLEMENTED))
+          {
+            tf_stream_error (stream, 0, myerror->message);
+            WARNING (stream, "Error setting codec preferences: %s",
+                myerror->message);
+            g_clear_error (&myerror);
+            return;
+          }
         g_clear_error (&myerror);
-        return;
       }
 
-
   if (g_object_class_find_property (
           G_OBJECT_GET_CLASS (stream->priv->fs_session),
           "no-rtcp-timeout"))
     g_object_set (stream->priv->fs_session, "no-rtcp-timeout", 0, NULL);
 
-  g_signal_connect (stream->priv->fs_stream, "src-pad-added",
-      G_CALLBACK (cb_fs_stream_src_pad_added), stream);
+  g_signal_connect_object (stream->priv->fs_stream, "src-pad-added",
+      G_CALLBACK (cb_fs_stream_src_pad_added), stream, 0);
 
   stream->priv->send_local_codecs = TRUE;
 
@@ -975,7 +1023,7 @@ cb_fs_local_candidates_prepared (TfStream *self)
               3, proto,
               4, "RTP",
               5, "AVP",
-              6, (double) candidate->priority / 65536.0,
+              6, (double) (candidate->priority / 65536.0),
               7, type,
               8, candidate->username,
               9, candidate->password,
@@ -1092,7 +1140,7 @@ tp_transports_to_fs (const gchar* foundation, const GPtrArray *transports)
           g_value_get_uint (g_value_array_get_nth (transport, 2))); /* port */
 
       fs_candidate->priority = (gint)
-          g_value_get_double (g_value_array_get_nth (transport, 6)) * 65536.0;
+          (g_value_get_double (g_value_array_get_nth (transport, 6)) * 65536.0);
       fs_candidate->username =
           g_value_dup_string (g_value_array_get_nth (transport, 8));
       fs_candidate->password =
@@ -1676,7 +1724,11 @@ cb_fs_new_active_candidate_pair (TfStream *self,
     FsCandidate *local_candidate,
     FsCandidate *remote_candidate)
 {
-  DEBUG (self, "called");
+  DEBUG (self, "called: c:%d local: %s %s:%u  remote: %s %s:%u",
+      local_candidate->component_id,
+      local_candidate->foundation, local_candidate->ip, local_candidate->port,
+      remote_candidate->foundation, remote_candidate->ip,
+      remote_candidate->port);
 
   tp_cli_media_stream_handler_call_new_active_candidate_pair (
     self->priv->stream_handler_proxy, -1, local_candidate->foundation,
@@ -1920,8 +1972,10 @@ cb_fs_stream_src_pad_added (FsStream *fsstream G_GNUC_UNUSED,
     gpointer user_data)
 {
   TfStream *self = TF_STREAM (user_data);
+  gchar *padname = gst_pad_get_name (pad);
 
-  DEBUG (self, "New pad");
+  DEBUG (self, "New pad %s: " FS_CODEC_FORMAT, padname, FS_CODEC_ARGS (codec));
+  g_free (padname);
 
   g_signal_emit (self, signals[SRC_PAD_ADDED], 0, pad, codec);
 }
